SYMCXX_TYPE(Symbol , NONE     , NONE   )
SYMCXX_TYPE(Integer, NONE     , NONE   )
SYMCXX_TYPE(MatProx, NONE     , NONE   )
SYMCXX_TYPE(Float  , NONE     , NONE   ) // All before (incl.) Float are atomic
