SYMCXX_TYPE(Sub    , Binary   , sub    )
SYMCXX_TYPE(Div    , Binary   , div    )
SYMCXX_TYPE(Add2   , Binary   , add2   )
SYMCXX_TYPE(Mul2   , Binary   , mul2   )
#include "symcxx/types_nonatomic_binary_math_h.inc"
#include "symcxx/types_nonatomic_binary_relational.inc"