SYMCXX_TYPE(Lt     , Binary   , lt     ) // All after (incl.) Lt are relational
SYMCXX_TYPE(Le     , Binary   , le     )
SYMCXX_TYPE(Eq     , Binary   , eq     )
SYMCXX_TYPE(Ne     , Binary   , ne     )
SYMCXX_TYPE(Ge     , Binary   , ge     )
SYMCXX_TYPE(Gt     , Binary   , gt     )
