SYMCXX_TYPE(Cos    , Unary    , cos    )
SYMCXX_TYPE(Sin    , Unary    , sin    )
SYMCXX_TYPE(Tan    , Unary    , tan    )
SYMCXX_TYPE(Acos   , Unary    , acos   )
SYMCXX_TYPE(Asin   , Unary    , asin   )
SYMCXX_TYPE(Atan   , Unary    , atan   )
SYMCXX_TYPE(Cosh   , Unary    , cosh   )
SYMCXX_TYPE(Sinh   , Unary    , sinh   )
SYMCXX_TYPE(Tanh   , Unary    , tanh   )
SYMCXX_TYPE(Acosh  , Unary    , acosh  )
SYMCXX_TYPE(Asinh  , Unary    , asinh  )
SYMCXX_TYPE(Atanh  , Unary    , atanh  )
SYMCXX_TYPE(Exp    , Unary    , exp    )
SYMCXX_TYPE(Log    , Unary    , log    )
SYMCXX_TYPE(Log10  , Unary    , log10  )
SYMCXX_TYPE(Exp2   , Unary    , exp2   )
SYMCXX_TYPE(Expm1  , Unary    , expm1  )
SYMCXX_TYPE(Log1p  , Unary    , log1p  )
SYMCXX_TYPE(Log2   , Unary    , log2   )
SYMCXX_TYPE(Logb   , Unary    , logb   )
SYMCXX_TYPE(Sqrt   , Unary    , sqrt   )
SYMCXX_TYPE(Cbrt   , Unary    , cbrt   )
SYMCXX_TYPE(Erf    , Unary    , erf    )
SYMCXX_TYPE(Erfc   , Unary    , erfc   )
SYMCXX_TYPE(Tgamma , Unary    , tgamma )
SYMCXX_TYPE(Lgamma , Unary    , lgamma )
