# -*- coding: utf-8 -*-
"""
symcxx is a minimal symbolic manipulation python package written in C++.
"""
from __future__ import (absolute_import, division, print_function)

from ._symcxx import PyNameSpace as NameSpace
from ._release import __version__
